
{-#  OPTIONS --type-in-type --without-K  #-}

module Preliminaries where

Type = Set

data Path {A : Type} : A → A → Type where
  id : {M : A} → Path M M

_≃_ : {A : Type} → A → A → Type
_≃_ = Path

infix 9 _≃_

path-induction : {A : Type} {M : A}
  (C : (x : A) → Path M x → Type) (b : C M id)
  {N : A}(α : Path M N) → C N α
path-induction _ b id = b

! : {A : Type} {M N : A} → Path M N → Path N M 
! id = id

_∘_  : {A : Type} {M N P : A} → Path N P → Path M N → Path M P
β ∘ id = β

infixr 10 _∘_ 

infix  2 _∎
infixr 2 _≃〈_〉_ 

_≃〈_〉_ : {A : Type} (x : A) {y z : A} → Path x y → Path y z → Path x z
_ ≃〈 p1 〉 p2 = (p2 ∘ p1)

_∎ : ∀ {A : Type} (x : A) → Path x x
_∎ _ = id

∘-unit-l  : {A : Type} {M N : A} (α : Path M N)
          → Path (id ∘ α) α
∘-unit-l id = id

∘-assoc  : {A : Type} {M N P Q : A} 
         (γ : Path P Q) (β : Path N P) (α : Path M N) 
         → Path (γ ∘ (β ∘ α)) ((γ ∘ β) ∘ α)
∘-assoc id id id = id

!-inv-l  : {A : Type} {M N : A} (α : Path M N) 
         → Path (! α ∘ α) id
!-inv-l id = id

∘-unit-r  : {A : Type} {M N : A} (α : Path M N)
          → Path (α ∘ id) α
∘-unit-r id = id

!-inv-r : {A : Type} {M N : A} (α : Path M N) → Path (α ∘ (! α)) id
!-inv-r id = id

!-inv-r-front : {A : Type} {M N P : A} (p : Path P N) (q : Path M N) → Path (p ∘ (! p) ∘ q) q
!-inv-r-front id id = id

!-inv-r-back : {A : Type} {M N P : A} (q : Path N M) (p : Path P N) → Path (q ∘ (p ∘ ! p)) q
!-inv-r-back id id = id

!-invol : {A : Type} {M N : A} (p : Path M N) → Path (! (! p)) p
!-invol id = id

!-∘ : {A : Type} {M N P : A} → (q : Path N P) → (p : Path M N)
    → (! (q ∘ p)) ≃ ! p ∘ ! q
!-∘ id id = id

move-right :  {A : Type} {M N P : A}
              (β : Path N P) (α : Path M N) (γ : Path M P)
           → Path (β ∘ α) γ
           → Path α (! β ∘ γ)
move-right id id γ x = ! (∘-unit-l γ) ∘ x

move-left-! :  {A : Type} {M N P : A}
              (α : Path M N) (β : Path N P) (γ : Path M P)
           → Path α (! β ∘ γ)
           → Path (β ∘ α) γ
move-left-! id id γ x = ∘-unit-l γ ∘ x

move-right-! :  {A : Type} {M N P : A}
               (β : Path P N) (α : Path M N) (γ : Path M P)
            → Path (! β ∘ α) γ
            → Path α (β ∘ γ)
move-right-! id id γ x = ! (∘-unit-l γ) ∘ x

move-right-right-! :  {A : Type} {M N P : A}
                      (β : Path N P) (α : Path N M) (γ : Path M P)
                   → Path (β ∘ ! α) γ
                   → Path β (γ ∘ α)
move-right-right-! id id γ x = x

rassoc-1-3-1 : ∀ {A} {M1 M2 M3 M4 M5 M6 : A}
               (a56 : Path M5 M6) (a45 : Path M4 M5) (a34 : Path M3 M4) (a23 : Path M2 M3) (a12 : Path M1 M2)
             → Path (a56 ∘ (a45 ∘ a34 ∘ a23) ∘ a12) (a56 ∘ a45 ∘ a34 ∘ a23 ∘ a12)
rassoc-1-3-1 id id id id id = id

transport :  {B : Type} (E : B → Type) 
             {b1 b2 : B} → Path b1 b2 → (E b1 → E b2)
transport C id = λ x → x

_o_ : {A B C : Type} → (B → C) → (A → B) → A → C
g o f = \ x → g (f x)
infixr 10 _o_

transport-Path-right :  {A : Type} {M N P : A} 
  (α' : Path N P) (α : Path M N)
  → Path (transport (\ x → Path M x) α' α) (α' ∘ α)

transport-→ :  {Γ : Type} (A B : Γ → Type) {θ1 θ2 : Γ} 
               (δ : θ1 ≃ θ2) (f : A θ1 → B θ1) 
  → Path  (transport (\ γ → (A γ) → B γ) δ f) 
          (transport B δ o f o (transport A (! δ)))

transport-Path-right id id = id 
transport-→ _ _ id f = id 



ap :  {A B : Type} {M N : A}
      (f : A → B) → Path{A} M N → Path{B} (f M) (f N)
ap f id = id

apd  :  {B : Type} {E : B → Type} {b₁ b₂ : B} 
        (f : (x : B) → E x) (β : Path b₁ b₂) 
     → Path (transport E β (f b₁)) (f b₂) 
apd f id = id 

transport-∘ : {A : Type} (C : A → Type) {M N P : A} (β : Path N P) (α : Path M N)
          → Path (transport C (β ∘ α)) (\ x → transport C β (transport C α x))
transport-∘ _ id id = id

transport-Path : {Γ A : Type} (f g : Γ → A) {M N : Γ} (p : Path M N)
               → (p' : _) → Path (transport (\ x → Path (f x) (g x)) p p')
                                   ((ap g p) ∘ p' ∘ (! (ap f p)))
transport-Path _ _ id p' = ! (∘-unit-l p')

transport-ap-assoc : {A : Type} (C : A → Type) {M N : A} (α : Path M N) → Path (transport C α) (transport (\ x → x) (ap C α))
transport-ap-assoc C id = id 

transport-ap-assoc' : {A B : Type} (C : B → Type) (f : A → B) {M N : A} (α : Path M N) → Path (transport (\ x -> C (f x)) α) (transport C (ap f α))
transport-ap-assoc' C f id = id 

ap-o : {A B C : Type} (g : B → C) (f : A → B)
           {M N : A} (α : Path M N)
         → ap (\ x → g (f x)) α ≃ ap g (ap f α)
ap-o g f id = id

ap-id : {A : Type} {M N : A} (α : Path M N)
       → Path (ap (\ x → x) α) α
ap-id id = id 

ap-! : {A B : Type} (F : A → B) {M N : A} (α : Path M N)
      → Path (ap F (! α)) (! (ap F α))
ap-! _ id = id 

ap-∘ : {A B : Type} (F : A → B) {M N P : A} (β : Path N P) (α : Path M N)
     → Path (ap F (β ∘ α)) (ap F β ∘ ap F α)
ap-∘ _ _ id = id

ap-by-id : {A : Type} {f : A → A}
               (αf : (x : _) → x ≃ f x) 
            → {M N : A} (α : M ≃ N)
            → (ap f α ≃ αf N ∘ α ∘ ! (αf M))
ap-by-id αf id = ap (λ x → αf _ ∘ x) (! (∘-unit-l (! (αf _)))) ∘ ! (!-inv-r (αf _)) 

ap≃ : ∀ {A} {B : A → Type} {f g : (x : A) → B x} 
      → Path f g → {x : A} → Path (f x) (g x)
ap≃ α {x} = ap (\ f → f x) α

postulate 
  λ≃  : ∀ {A} {B : A → Type} {f g : (x : A) → B x}
      → ((x : A) → Path (f x) (g x)) 
      → Path f g

record HEquiv (A B : Type) : Type where
  constructor hequiv
  field
   f : A → B
   g : B → A
   α : (x : A) → Path (g (f x)) x
   β : (y : B) → Path (f (g y)) y

!-equiv : ∀ {A B} → HEquiv A B → HEquiv B A
!-equiv (hequiv f g α β) = hequiv g f β α

postulate

 univalence : {A B : Type} → HEquiv A B → Path A B

 transport-univ : {A B : Type} (e : HEquiv A B) 
  → Path  (transport (\ (A : Type) → A) (univalence e))
          (HEquiv.f e)
  
 !-univalence : {A B : Type} (e : HEquiv A B) 
              → Path  (! (univalence e))
                      (univalence (!-equiv e))

data Positive : Type where
  One : Positive
  S   : (n : Positive) → Positive

data Int : Type where
  Pos  : (n : Positive) → Int
  Zero : Int
  Neg  : (n : Positive) → Int


succ : Int → Int
succ Zero = Pos One
succ (Pos n) = Pos (S n)
succ (Neg One) = Zero
succ (Neg (S n)) = Neg n

pred : Int → Int
pred Zero = Neg One
pred (Pos (S n)) = Pos n
pred (Pos One) = Zero
pred (Neg n) = Neg (S n)

_+_ : Int → Int → Int
Zero + m        = m
(Pos One) + m   = succ m
(Pos (S n)) + m = succ (Pos n  +  m)
(Neg One)   + m = pred m
(Neg (S n)) + m = pred (Neg n  +  m)

pred-succ : (n : Int) → Path (pred (succ n)) n
pred-succ (Pos y) = id
pred-succ (Zero) = id
pred-succ (Neg One) = id
pred-succ (Neg (S y)) = id

succ-pred : (n : Int) → Path (succ (pred n)) n
succ-pred (Pos One) = id
succ-pred (Pos (S y)) = id
succ-pred (Zero) = id
succ-pred (Neg y) = id

succEquiv : HEquiv Int Int
succEquiv = hequiv succ pred pred-succ succ-pred
